#!/usr/bin/env python
import sys
import os
import functools
import itertools
import bisect

import healpy

import matplotlib
matplotlib.use("Agg")
from matplotlib import pyplot
from mpl_toolkits.basemap import Basemap

import numpy
import scipy.integrate
import scipy.special

from lalinference.bayestar import fits as bfits
from lalinference.bayestar import plot as bplot

import RIFT.integrators.mcsampler as mcsampler
from statutils import cumvar


__author__ = "Chris Pankow <pankow@gravity.phys.uwm.edu>"

#
# FITS utilities
#
def make_skymap(samp, res=32, fname=None):
	"""
	Plot the RA and dec distribution in a Mollewide projection.
	"""
	samples = samp._rvs
	smap = numpy.zeros(healpy.nside2npix(res))
	phi, thetar = samples[("ra", "dec")]
	# FIXME: Why is this flipped?
	for ind in healpy.ang2pix(res, numpy.pi - (thetar + numpy.pi/2), numpy.pi*2 - phi):
		smap[ind] += 1
	bfits.write_sky_map(fname or "pe_skymap.fits", smap)


#
# Plotting utilities
#
def plot_integral(fcn, samp, fargs=None, fname=None):
	"""
	Effectively redoes the process which happens in MCSampler.integrate, but keeps track of the values for plotting.
	"""
	args = []
	p_s = numpy.array([])
	for p in fargs or samp.params:
		if isinstance(p, tuple) and len(p) > 1:
			# FIXME: Will this preserve order?
			for a in samp._rvs[p]:
				args.append(a)
			p_s = numpy.hstack( (p_s, samp.pdf[p](*samp._rvs[p])/samp._pdf_norm[p]) )
		else:
			p_s = numpy.hstack( (p_s, samp.pdf[p](samp._rvs[p])/samp._pdf_norm[p]) )
			args.append(samp._rvs[p])

	p_s.shape = (len(samp.params), len(p_s)/len(samp.params))

	fval = fcn(*args)
	n = len(fval)
	#p_s = numpy.array([ samp.pdf[p](*x)/samp._pdf_norm[p] for p, x in samp._rvs.items() ])
	joint_p_s = numpy.prod(p_s, axis=0)
	int_val = fval/joint_p_s
	maxval = [fval[0]/joint_p_s[0] or -float("Inf")]
	for v in int_val[1:]:
			maxval.append( v if v > maxval[-1] and v != 0 else maxval[-1] )
	eff_samp = int_val.cumsum()/maxval

	pyplot.figure()
	pyplot.subplot(311)
	pyplot.title("Integral estimate")
	n_itr = range(1, len(int_val)+1)
	pyplot.semilogx()
	pyplot.plot(n_itr, int_val.cumsum()/numpy.linspace(1,n,n), 'k-')
	pyplot.ylabel("integral val")
	pyplot.twinx()
	pyplot.ylabel("integral std")
	pyplot.plot(n_itr, numpy.sqrt(cumvar(int_val)/n_itr), 'b-')
	pyplot.grid()
	pyplot.subplot(312)
	pyplot.title("Point integral estimate and maximum over iterations")
	pyplot.plot(n_itr, maxval, 'k-')
	pyplot.plot(n_itr, int_val, 'b-')
	pyplot.grid()
	pyplot.semilogx()
	pyplot.subplot(313)
	pyplot.title("Effective samples")
	pyplot.loglog()
	pyplot.ylabel("N_eff")
	pyplot.plot(range(1, len(int_val)+1), eff_samp, 'k-')
	pyplot.twinx()
	pyplot.loglog()
	pyplot.ylabel("ratio N_eff/N")
	pyplot.plot(range(1, len(int_val)+1), eff_samp/numpy.arange(1,len(int_val)+1), 'b-')
	pyplot.grid()
	pyplot.subplots_adjust(hspace=0.5)
	pyplot.savefig(fname or "integral.pdf")
	pyplot.close()

def plot_integrand(fcn, x1, x2, fname=None):
	"""
	Plot 1D integrand (fcn) from x1 to x2
	"""
	pyplot.figure()
	x_i = numpy.linspace(x1, x2, 1000)
	pyplot.plot(x_i, fcn(x_i))
	pyplot.savefig(fname or "integrand.pdf")
	pyplot.close()


def plot_pdf(samp, fname=None):
	"""
	Plot PDFs of sampling distributions if they are independent
	"""
	pyplot.figure()
	for param in samp.params:
		x_i = numpy.array([numpy.linspace(samp.llim[param], samp.rlim[param], 100)])
		pyplot.plot(x_i[0], map(samp.pdf[param], x_i)[0], '-', label=param)

	pyplot.grid()
	pyplot.legend()
	pyplot.savefig(fname or "pdf.pdf")
	pyplot.close()

def plot_cdf_inv(samp, fname=None):
	"""
	Plot inverse CDFs
	"""

	pyplot.figure()
	x_i = numpy.linspace(0, 1, 1000)
	for param, cdfinv in samp.cdf_inv.items():
		pyplot.plot(x_i, map(cdfinv, x_i), '-', label=param)

	pyplot.grid()
	pyplot.legend()
	pyplot.savefig(fname or "cdf_inv.pdf")
	pyplot.close()


def plot_one_d_hist(samp, fname=None):
	"""
	Plot one-d histograms of all parameters
	"""
	fig = pyplot.figure(figsize=(10,4))
	samples = samp._rvs
	np = len(samp.params)
	i=1
	for p in samp.params:
		subp = pyplot.subplot(1, np, i)
		i += 1
		pyplot.hist(samples[p], bins=20)
		pyplot.grid()
		pyplot.xlabel(p)
	pyplot.savefig(fname or "samples_1d.pdf")
	pyplot.close()

def plot_two_d_hist(samp, fname=None):
	"""
	Plot two-d histograms of all parameters
	"""
	samples = samp._rvs
	fig = pyplot.figure(figsize=(10,10))
	np = len(samp.params)-1
	i, j = 1, 1
	label_next = True
	for (p1, p2) in itertools.combinations(samp.params, 2):
		subp = pyplot.subplot(np, np, i)
		s1, s2 = samples[p1], samples[p2]
		hist, yedge, xedge = numpy.histogram2d(s2, s1, bins=(100, 100))
	
		pyplot.imshow(hist.T, extent=(yedge[0], yedge[-1], xedge[-1], xedge[0]), interpolation='nearest')
		im = subp.get_images()
		ex = im[0].get_extent()
		subp.set_aspect(abs((ex[1]-ex[0])/(ex[3]-ex[2])))
		pyplot.colorbar()
		if label_next:
			pyplot.ylabel(p1)
			pyplot.xlabel(p2)
			label_next = False
	
		if i % np == 0:
			i += j
			j += 1
			label_next = True
		i += 1
	
	pyplot.savefig(fname or "samples_2d.pdf", figsize=(10,3))
	pyplot.close()
	
def plot_ra_dec(samp, use_pdf=False, fname=None, key=("ra", "dec")):
	"""
	Plot the RA and dec distribution in a Mollewide projection.
	"""
	samples = samp._rvs
	if isinstance(key, tuple):
		ra_samp, dec_samp = samples[key]
		ra_min, dec_min = samp.llim[key]
		ra_max, dec_max = samp.rlim[key]
		pdf = samp.pdf[key]
	else:
		ra_samp, dec_samp = samples[key[0]], samples[key[1]]
		ra_min, ra_max = samp.llim[key[0]], samp.rlim[key[0]]
		dec_min, dec_max = samp.llim[key[1]], samp.rlim[key[1]]
		def radec(ra, dec):
			return (samp.pdf[key[0]], samp.pdf[key[1]])
		pdf = radec

	fig = pyplot.figure(figsize=(10,10))
	m = Basemap(projection='moll', lon_0=0, resolution='c')
	if not use_pdf:
		hist, xedge, yedge = numpy.histogram2d(ra_samp, dec_samp, bins=(100, 100), normed=True)
		x, y = numpy.meshgrid(xedge, yedge)
		x *= 180/numpy.pi
		y *= 180/numpy.pi
		m.contourf(x[:-1,:-1], y[:-1,:-1], hist.T, 100, cmap=matplotlib.cm.jet, latlon=True)
	else:
		xedge = numpy.linspace(ra_min, ra_max, 100)
		yedge = numpy.linspace(dec_min, dec_max, 100)
		x, y = numpy.meshgrid(xedge, yedge)
		hist = pdf(x, y)
		x *= 180/numpy.pi
		y *= 180/numpy.pi
		m.contourf(x, y, hist, cmap=matplotlib.cm.jet, latlon=True)

	pyplot.colorbar()
	m.drawparallels(numpy.arange(-90, 120, 30))
	m.drawmeridians(numpy.arange(0, 420, 60))
	pyplot.savefig(fname or "radec_proj.pdf")
	pyplot.close()
	return hist

#
# set up bounds on parameters
#

# Polarization angle
#psi_min, psi_max = 0, 2*numpy.pi
psi_min, psi_max = -2*numpy.pi, 2*numpy.pi
psi_val, psi_width = numpy.pi/2, numpy.pi/6
# RA and dec
ra_min, ra_max = 0, 2*numpy.pi
#ra_min, ra_max = -2*numpy.pi, 2*numpy.pi
ra_val, ra_width = numpy.pi/4, 5*numpy.pi/180
dec_min, dec_max = -numpy.pi/2, numpy.pi/2
dec_val, dec_width = -numpy.pi/4, 5*numpy.pi/180
# Inclination angle
inc_min, inc_max = -numpy.pi/2, numpy.pi/2
inc_val, inc_width = -numpy.pi/3, 10*numpy.pi/180
# orbital phi
phi_min, phi_max = 0, 2*numpy.pi
phi_val, phi_width = numpy.pi/5, 10*numpy.pi/180
# distance
dist_min, dist_max = 0.0, 100.0
dist_val, dist_width = 25.0, 25.0


samp = mcsampler.MCSampler()

#
# Test some other 1-D integrals
#
"""
test_min, test_max = -10, 10
def integrand_1d(x):
	return numpy.sinc(10*x)
plot_integrand(integrand_1d, test_min, test_max)

samp.add_parameter("p1", functools.partial(mcsampler.uniform_samp_vector, test_min, test_max), None, test_min, test_max)
plot_cdf_inv(samp)
res, var = samp.integrate(integrand_1d, int(sys.argv[1]), "p1")
print res, var
integral = scipy.integrate.quad(integrand_1d, psi_min, psi_max)[0]
print integral
samp.clear()
"""

def integrand_1d(p):
	return 1.0/numpy.sqrt(2*numpy.pi*psi_width**2)*numpy.exp(-(p-psi_val)**2/2.0/psi_width**2)
plot_integrand(integrand_1d, psi_min, psi_max)

#
# Test 1: What happens when we know the exact right answer
#

d = "test1/"
if not os.path.exists(d): os.makedirs(d)
os.chdir(d)

samp.add_parameter("psi", functools.partial(mcsampler.gauss_samp, psi_val, psi_width), None, psi_min, psi_max)
sampler_integral = samp.integrate(integrand_1d, "psi", nmax=1)
integral = scipy.integrate.quad(integrand_1d, psi_min, psi_max)[0]
print("scipy answer vs our answer:",  integral, sampler_integral)
plot_pdf(samp)
plot_cdf_inv(samp)
plot_one_d_hist(samp)
samp.clear()

os.chdir("../")

#
# Test 2: Same integrand, uniform sampling
#

d = "test2/"
if not os.path.exists(d): os.makedirs(d)
os.chdir(d)

samp.add_parameter("psi", functools.partial(mcsampler.uniform_samp_vector, psi_min, psi_max), None, psi_min, psi_max)
#print samp.integrate(integrand_1d, "psi", neff=1000, nmax=int(sys.argv[1]))
res, var = samp.integrate(integrand_1d, "psi", neff=1000)
print("Integral: %f, stddev: %f" % (res, numpy.sqrt(var)))
plot_integral(integrand_1d, samp)
plot_pdf(samp)
plot_cdf_inv(samp)
plot_one_d_hist(samp)
samp.clear()

os.chdir("../")

#
# Test 3: Same integrand, gaussian sampling -- various width and offsets
#
widths = [0.1, 0.2, 0.5, 0.8, 0.9, 0.99, 1.01, 1.1, 1.5, 2, 4, 5, 10]
offsets = [-2.0, -1.0, -0.5, -0.25, -0.1, -0.01, 0.01, 0.1, 0.25, 0.5, 1.0, 2.0]

variances = []
wo = []
d = "test3/"
if not os.path.exists(d): os.makedirs(d)
for w, o in itertools.product(widths, offsets):
	d1 = "test3/" + "w_%2.2f_o_%2.2f/" % (w,o)
	if not os.path.exists(d1): os.makedirs(d1)
	os.chdir(d1)
	samp.add_parameter("psi", functools.partial(mcsampler.gauss_samp, psi_val-o*psi_width, w*psi_width), None, psi_min, psi_max)
	plot_pdf(samp)
	plot_cdf_inv(samp)
	print("Width of Gaussian sampler (in units of the width of the integrand: %f" % w)
	print("offset of Gaussian sampler (in units of the width of the integrand: %f" % o)
	res, var = samp.integrate(integrand_1d, "psi", neff=100, nmax=int(1e5))
	plot_integral(integrand_1d, samp)
	variances.append(var)
	wo.append((w, o))
	print("Integral: %f, stddev: %f" % (res, numpy.sqrt(var)))
	samp.clear()
	os.chdir("../../")

pyplot.figure()
widths, offsets = zip(*wo)
pyplot.grid()
pyplot.xlabel("width in units of $\\sigma_{\\psi}$")
pyplot.ylabel("offset in units of $\\sigma_{\\psi}$")
pyplot.scatter(widths, offsets, c=numpy.log10(numpy.array(variances)))
cbar = pyplot.colorbar()
cbar.set_label("log10 variance")
pyplot.semilogx()
pyplot.savefig("gsamp_variances.pdf")

#
# Testing variance under convergence: Loop over samples and test sigma relation 
# for desired error
# 
pyplot.figure()

d = "test3b/"
if not os.path.exists(d): os.makedirs(d)
os.chdir(d)

samp.add_parameter("psi", functools.partial(mcsampler.uniform_samp_vector, psi_min, psi_max), None, psi_min, psi_max)
for n in 10**(numpy.arange(1,6)):
	ans = []
	print("Number of samples in integral %d" % n)
	for i in range(1000):
		res, var = samp.integrate(integrand_1d, "psi", nmax=n)
		ans.append( (res-1.0)/numpy.sqrt(var) )
	pyplot.hist(ans, bins=20, label="%d" % n)
pyplot.title("$(I-\\bar{I})/\\bar\\sigma$")
pyplot.grid()
pyplot.legend()
pyplot.savefig("integral_hist.pdf")
samp.clear()
exit()

os.chdir("../")

#
# Test 4: 2D PDFs, rejection sampling, and BAYESTAR
#

#
# Read FITS data
#
smap, smap_meta = bfits.read_sky_map("data/30602.toa.fits.gz")
sides = healpy.npix2nside(len(smap))

d = "test4/"
if not os.path.exists(d): os.makedirs(d)
os.chdir(d)

#
# For sampling PDF, have the option to apply a temperature argument
#
def bayestar_temp(temp, skymap, lon, lat):
	return bplot._healpix_lookup(skymap, lon, lat)**(1.0/temp)
def bayestar_norm(skymap, lon, lat):
	return bplot._healpix_lookup(skymap, lon, lat)*numpy.cos(lat)

sky_pdf = functools.partial(bayestar_temp, 1, smap)
norm_sky_pdf = functools.partial(bayestar_norm, smap)

#
# Renormalize for the integral
#
#norm = scipy.integrate.dblquad(sky_pdf, dec_min, dec_max, lambda x: ra_min, lambda x: ra_max, epsabs=1e-3, epsrel=1e-3)[0]
norm = scipy.integrate.dblquad(norm_sky_pdf, dec_min, dec_max, lambda x: ra_min, lambda x: ra_max, epsabs=1e-3, epsrel=1e-3)[0]
# FIXME: This shouldn't address variables that are "hidden"
samp._pdf_norm[("ra", "dec")] = norm

#
# Try to integrate the sky PDF itself
#
pix_radsq = len(smap)/4.0/numpy.pi
def integrand_2d(ra, dec):
	#return bayestar_temp(1, smap, ra, dec) * pix_radsq
	return numpy.ones(ra.shape)

generate_sky_points = functools.partial(mcsampler.sky_rejection, smap)
samp.add_parameter(("ra", "dec"), sky_pdf, generate_sky_points, (ra_min, dec_min), (ra_max, dec_max))

res, var = samp.integrate(integrand_2d, ("ra", "dec"), neff=1000)
print("Integral: %f, stddev: %f" % (res, numpy.sqrt(var)))
plot_ra_dec(samp)
plot_ra_dec(samp, use_pdf=True, fname="pdf.pdf")
plot_integral(integrand_2d, samp)
make_skymap(samp)

# FIXME: Never converges
#print "scipy says %g" % scipy.integrate.dblquad(integrand_2d, dec_min, dec_max, lambda x: ra_min, lambda x: ra_max, epsabs=1e-3, epsrel=1e-3)[0]

samp.clear()
os.chdir("../")

#
# Test 5a: Uniform sampling, cdf provided
#
d = "test5a/"
if not os.path.exists(d): os.makedirs(d)
os.chdir(d)

#samp.add_parameter("psi", functools.partial(mcsampler.uniform_samp_vector, psi_min, psi_max), functools.partial(mcsampler.inv_uniform_cdf, psi_min, psi_max), psi_min, psi_max)
#samp.add_parameter("ra", functools.partial(mcsampler.uniform_samp_vector, ra_min, ra_max), functools.partial(mcsampler.inv_uniform_cdf, ra_min, ra_max), ra_min, ra_max)
#samp.add_parameter("dec", functools.partial(mcsampler.uniform_samp_vector, dec_min, dec_max), functools.partial(mcsampler.inv_uniform_cdf, dec_min, dec_max), dec_min, dec_max)

#
# Test 5b: Uniform sampling, cdf provided
#
# Uniform sampling, auto-cdf inverse
#samp.add_parameter("psi", functools.partial(mcsampler.uniform_samp_vector, psi_min, psi_max), None, psi_min, psi_max)
samp.add_parameter("psi", functools.partial(mcsampler.gauss_samp, psi_val, 2*psi_width), None, psi_min, psi_max)
samp.add_parameter("ra", functools.partial(mcsampler.uniform_samp_vector, ra_min, ra_max), None, ra_min, ra_max)
samp.add_parameter("dec", functools.partial(mcsampler.uniform_samp_vector, dec_min, dec_max), None, dec_min, dec_max)
samp.add_parameter("phi", functools.partial(mcsampler.uniform_samp_vector, phi_min, phi_max), None, phi_min, phi_max)
#samp.add_parameter("inc", functools.partial(mcsampler.uniform_samp_vector, inc_min, inc_max), None, inc_min, inc_max)
samp.add_parameter("inc", functools.partial(mcsampler.gauss_samp, inc_val, 2*inc_width), None, inc_min, inc_max)
samp.add_parameter("dist", functools.partial(mcsampler.uniform_samp_vector, dist_min, dist_max), None, dist_min, dist_max)

#
# Test 5c: Gaussian sampling, auto-cdf inverse -- Doesn't work yet
#
#samp.add_parameter("psi", functools.partial(mcsampler.gauss_samp, 0, (psi_max-psi_min)/3.0), None, psi_min, psi_max)
#samp.add_parameter("ra", functools.partial(mcsampler.gauss_samp, 0, (ra_max-ra_min)/10.0), None, ra_min, ra_max)
#samp.add_parameter("dec", functools.partial(mcsampler.gauss_samp, (dec_max+dec_min)/2, (dec_max-dec_min)/10.0), None, dec_min, dec_max)

#
# Full 6-D test
#

a, b, c, d, e, f = 2*psi_width**2, 2*ra_width**2, 2*dec_width**2, 2*inc_width**2, 2*phi_width**2, 2*dist_width**2
norm = 1.0/numpy.sqrt((numpy.pi)**len(samp.params)*a*b*c*d*e*f)
numpy.seterr(invalid="raise")
def integrand(p, r, dec, ph, i, di):
	exponent = -(p-psi_val)**2/a-(r-ra_val)**2/b-(dec-dec_val)**2/c-(i-inc_val)**2/d-(ph-phi_val)**2/e-(di-dist_val)**2/f
	return norm * numpy.exp(exponent)

res, var = samp.integrate(integrand, "psi", "ra", "dec", "phi", "inc", "dist", neff=100, nmax=int(1e6))
print("Integral value: %f, stddev %f" % (res, numpy.sqrt(var)))

plot_integral(integrand, samp, ("psi", "ra", "dec", "phi", "inc", "dist"))
plot_pdf(samp)
plot_cdf_inv(samp)
plot_one_d_hist(samp)
plot_two_d_hist(samp)
plot_ra_dec(samp, key=["ra", "dec"])

os.chdir("../")



#! /usr/bin/env python
from __future__ import division, print_function


def get_args(raw_args):
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "channel",
        help="Channel to add.",
    )

    parser.add_argument(
        "input_frame",
        nargs="+",
        help="One of the frame files to be added.",
    )

    parser.add_argument(
        "output_frame",
        help="Frame file to store result in.",
    )

    return parser.parse_args(raw_args)


def main(raw_args):
    args = get_args(raw_args)

    import pycbc.frame


    output_timeseries = None
    print(args.input_frame)
    for frame in args.input_frame:
        print(frame, args.channel)
        timeseries = pycbc.frame.read_frame(
            frame, args.channel
        )

        start_time = timeseries.start_time
        delta_t = timeseries.delta_t
        end_time = timeseries.end_time
        print("Start Time: ",start_time)
        print("End Time: ",end_time)
        # A bit of a hack.
        # I've encountered frames that appear to have their first
        # sample discarded. As a result, the start_time is delta_t
        # greater than an integer value. So we take the fractional
        # part of the start time (t - int(t)) and compare to the
        # timestep delta_t. If they're the same, we've hit that case,
        # and we simply prepend a zero to the timeseries, making it
        # start at an integer time.
        if delta_t == (start_time - int(start_time)):
            print("Doing Hack (start)")
            timeseries.prepend_zeros(1)
        if  abs(end_time - int(end_time)) > 0:
            print("Doing Hack (end)")
            timeseries.append_zeros(1)
        print(timeseries,len(timeseries))

        if output_timeseries is None:
            output_timeseries = timeseries
        else:
            output_timeseries += timeseries

    pycbc.frame.write_frame(
        args.output_frame,
        args.channel,
        output_timeseries,
    )


if __name__ == "__main__":
    import sys
    raw_args = sys.argv[1:]
    sys.exit(main(raw_args))



# Test script for comparing GMM integrator to existing mcsampler integrator in
# RIFT. A simple n-dimensional integrand consisting of a highly-correlated
# Gaussian is used.

from __future__ import print_function
import numpy as np
from scipy.stats import multivariate_normal
from scipy.stats import truncnorm
import matplotlib.pyplot as plt

from RIFT.integrators import mcsampler, mcsamplerEnsemble

### test parameters

# width of domain of integration, same for all dimensions
width = 10.0                                                    
# number of dimensions
ndim = 3                                                        
# mean of the Gaussian, allowed to occupy middle half of each dimension
mu = np.random.uniform(-1 * width / 4.0, width / 4.0, ndim)    
# number of iterations for mcsamplerEnsemble
n_iters = 40                                                    
# max number of samples for mcsampler
nmax = 40000                                                    

llim = -1 * width / 2
rlim = width / 2

### generate list of named parameters
params = [str(i) for i in range(ndim)]

### generate the covariance matrix
cov = np.identity(ndim)
cov[ndim - 1][ndim - 1] = 0.05 # make it narrower in one dimension

### add some covariance (to test handling of strongly-correlated likelihoods)
cov[0][ndim - 1] = -0.1
cov[ndim - 1][0] = -0.1

### define integrand as a weighted sum of Gaussians
def f(x1, x2, x3):
    x = np.array([x1, x2, x3]).T
    return (multivariate_normal.pdf(x, mu, cov))

### initialize samplers
sampler = mcsampler.MCSampler()
samplerEnsemble = mcsamplerEnsemble.MCSampler()

### add parameters
for p in params:
    sampler.add_parameter(p, np.vectorize(lambda x:1), 
            prior_pdf=np.vectorize(lambda x:1),
            left_limit=llim, right_limit=rlim,
            adaptive_sampling=True)
    samplerEnsemble.add_parameter(p, left_limit=llim, right_limit=rlim,adaptive_sampling=True)

# number of Gaussian components to use in GMM
n_comp = 1

### integrate
integral_1, var_1, eff_samp_1, _ = sampler.integrate(f, *params, 
        no_protect_names=True, nmax=20000, save_intg=True)
print(" --- finished default --")
integral_2, var_2, eff_samp_2, _ = samplerEnsemble.integrate(f, *params, 
        min_iter=n_iters, max_iter=n_iters, correlate_all_dims=True, n_comp=n_comp,super_verbose=True)
print(" --- finished GMM --")
print("mu",mu)
### CDFs

### get our posterior samples as a single array
arr_1 = np.empty((len(sampler._rvs["0"]), ndim))
arr_2 = np.empty((len(samplerEnsemble._rvs["0"]), ndim))
for i in range(ndim):
    arr_1[:,i] = sampler._rvs[str(i)].flatten()
    arr_2[:,i] = samplerEnsemble._rvs[str(i)].flatten()

colors = ["black", "red", "blue", "green", "orange"]

plt.figure(figsize=(10, 8))

for i in range(ndim):
    s = np.sqrt(cov[i][i])
    ### get sorted samples (for the current dimension)
    x_1 = arr_1[:,i][np.argsort(arr_1[:,i])]
    x_2 = arr_2[:,i][np.argsort(arr_2[:,i])]
    ### plot true cdf
    plt.plot(x_1, truncnorm.cdf(x_1, llim, rlim, mu[i], s), label="True CDF",
            color=colors[i], linewidth=0.5)
    # NOTE: old mcsampler stores L, mcsamplerEnsemble stores lnL
    L = sampler._rvs["integrand"]
    p = sampler._rvs["joint_prior"]
    ps = sampler._rvs["joint_s_prior"]
    ### compute weights of samples
    weights_1 = (L * p / ps)[np.argsort(arr_1[:,i])]
    L = samplerEnsemble._rvs["integrand"]
    p = samplerEnsemble._rvs["joint_prior"]
    ps = samplerEnsemble._rvs["joint_s_prior"]
    ### compute weights of samples
    weights_2 = (L * p / ps)[np.argsort(arr_2[:,i])]
    y_1 = np.cumsum(weights_1)
    y_1 /= y_1[-1] # normalize
    y_2 = np.cumsum(weights_2)
    y_2 /= y_2[-1] # normalize
    ### plot recovered cdf
    plt.plot(x_1, y_1, "--", label="Recovered CDF, mcsampler", color=colors[i],
            linewidth=2)
    plt.plot(x_2, y_2, ":", label="Recovered CDF, mcsamplerEnsemble",
            color=colors[i], linewidth=2)

plt.legend()

fname = "cdf.png"

print("Saving CDF figure as " + fname + "...")

plt.savefig(fname)



#! /usr/bin/env python
"""
util_PrintInjection --inj [inj]  # prints standard dump of properties
util_PrintInjection --inj [inj]  --psiJ --psiL --iota --m1  # print rows of these properties
"""

import argparse
import numpy as np
import lalsimulation as lalsim
import RIFT.lalsimutils as lalsimutils
import lal
import sys
from igwn_ligolw import utils, lsctables, table, ligolw

cthdler = lalsimutils.cthdler  #ligolw.LIGOLWContentHandler #defines a content handler to load xml grids
lsctables.use_in(cthdler)


parser = argparse.ArgumentParser()
parser.add_argument("--inj", dest='inj', default=None,help="inspiral XML file containing injection information.")
parser.add_argument("--event",type=int, dest="event_id", default=None,help="event ID of injection XML to use. If not specified loops over all")
parser.add_argument("--verbose", action="store_true",default=False)
parser.add_argument("--distance",action="store_true",default=False)
parser.add_argument("--right-ascension",action="store_true",default=False)
parser.add_argument("--declination",action="store_true",default=False)
parser.add_argument("--psiJ",action="store_true",default=False)
parser.add_argument("--psiL",action="store_true",default=False)
parser.add_argument("--iota",action="store_true",default=False)
parser.add_argument("--thetaJL",action="store_true",default=False)
parser.add_argument("--thetaJN",action="store_true",default=False)
parser.add_argument("--alpha",action="store_true",default=False)
parser.add_argument("--system-frame",action="store_true",default=True,help="If verbose, show system frame")
parser.add_argument("--force-morphology",type=int,default=3,help="If -1,0,1, select only that morphology, otherwise all")
parser.add_argument("--force-J-limit",type=float,default=-1,help="If a number, selects injections with J/M^2 less than it.")
parser.add_argument("--fref",type=float, default=0.0,help="For injections being interpreted at a different reference frequency than actually used in the file (historical)")
parser.add_argument("--show-fend",default=False,action='store_true')
opts=  parser.parse_args()

filename = opts.inj
event = opts.event_id
xmldoc = utils.load_filename(filename,contenthandler = cthdler, verbose = True)
Jlim = opts.force_J_limit


sim_inspiral_table = lsctables.SimInspiralTable.get_table(xmldoc)
if event is not None:
    events = [event]
else:
    events = np.arange( len(sim_inspiral_table))

P=lalsimutils.ChooseWaveformParams()
for event in events:
   P.copy_sim_inspiral(sim_inspiral_table[int(event)])
   if opts.fref:
       P.fref = opts.fref
   Jvec = P.TotalAngularMomentumAtReferenceOverM2()

   toprint = opts.verbose
   toprintHead = opts.verbose
   if not (opts.force_morphology == -1 or opts.force_morphology == 0  or opts.force_morphology == 1):
       toprint = opts.verbose
   else:
       if (opts.force_morphology == P.EvaluateMorphology()):
           toprintHead=True
       else:
           toprint = False
           
   if not( Jlim == -1) and np.sqrt(Jvec*Jvec)>Jlim:
       toprint=False
   else:
       toprintHead=True
   

       
   if toprint:
       if toprintHead:
                      print(" ----   Event number ", event , " ------")
       P.print_params(show_system_frame=opts.system_frame) # show_morphology=True, ,show_fend=opts.show_fend

 



import sys
import math
#import bisect
from collections import defaultdict
from types import ModuleType

import numpy
np=numpy #import numpy as np
from scipy import integrate, interpolate, special
import itertools
import functools

from copy import deepcopy


import os

try:
  import cupy
  import cupyx   # needed for logsumexp
  xpy_default=cupy
  try:
    xpy_special_default = cupyx.scipy.special
    if not(hasattr(xpy_special_default,'logsumexp')):
          print(" mcsamplerPortfolio no cupyx.scipy.special.logsumexp, fallback mode ...")
          xpy_special_default= special
  except:
    print(" mcsamplerPortfolio no cupyx.scipy.special, fallback mode ...")
    xpy_special_default= special
  identity_convert = cupy.asnumpy
  identity_convert_togpu = cupy.asarray
  junk_to_check_installed = cupy.array(5)  # this will fail if GPU not installed correctly
  cupy_ok = True
  cupy_pi = cupy.array(np.pi)

  from RIFT.interpolators.interp_gpu import interp

#  from logging import info as log
#  import inspect
#  def verbose_cupy_asarray(*args, **kwargs):
#     print("Transferring data to VRAM", *args, **kwargs)
#     return cupy.asarray(*args, **kwargs)
#  def verbose_cupy_asnumpy(*args, **kwargs):
#     curframe = inspect.currentframe()
#     calframe = inspect.getouterframes(curframe, 2)
#     log("Transferring data to RAM",calframe[1][3]) #,args[0].__name__) #, *args, **kwargs)
#     return cupy.ndarray.asnumpy(*args, **kwargs)
#  cupy.asarray = verbose_cupy_asarray  
#  cupy.ndarray.asnumpy = verbose_cupy_asnumpy

except:
  print(' no cupy (mcsamplerPortfolio)')
#  import numpy as cupy  # will automatically replace cupy calls with numpy!
  xpy_default=numpy  # just in case, to make replacement clear and to enable override
  xpy_special_default = special
  identity_convert = lambda x: x  # trivial return itself
  identity_convert_togpu = lambda x: x
  cupy_ok = False
  cupy_pi = np.pi

def set_xpy_to_numpy():
   xpy_default=numpy
   identity_convert = lambda x: x  # trivial return itself
   identity_convert_togpu = lambda x: x
   cupy_ok = False
   

if 'PROFILE' not in os.environ:
   def profile(fn):
        return fn


from RIFT.integrators.statutils import  update,finalize, init_log,update_log,finalize_log

#from multiprocessing import Pool

from RIFT.likelihood import vectorized_general_tools

# import matching integrators registered through plutings
#  https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points
discovered_plugins = entry_points(group='RIFT.integrator_plugins')
known_pipelines = {}
for pipeline in discovered_plugins:
  print(" Portfolio discovery: loading ", pipeline.name)
  known_pipelines[pipeline.name] = pipeline.load()
print('RIFT portfolio plugins:', [ep.name for ep in discovered_plugins])


class NanOrInf(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


###
### scheduling functions : return probability array, given 
###

def portfolio_default_weights(n_ess_list, wt_previous, portfolio_probability_floor=0.01, history_factor=0.5, xpy=xpy_default, identity_convert=lambda x:x, **kwargs):
  assert len(n_ess_list) == len(wt_previous)

  vals = n_ess_list
  rewt = vals - 1   # will be non-negative
  # don't update if we have insane answers
  if any(np.isnan(rewt)):
    return wt_previous
  rewt = np.ones(len(rewt))*portfolio_probability_floor + (rewt/np.sum(rewt)) * (1-portfolio_probability_floor)
  net = (rewt * history_factor + wt_previous*(1-history_factor))
  return net/np.sum(net) # make SURE normalized correctly


###
### PORTFOLIO CLASS
###


class MCSampler(object):
    """
    Class to define a set of parameter names, limits, and probability densities.
    """



    def __init__(self,portfolio=None,portfolio_weights=None,oracle_realizations =None,n_chunk=400000, portfolio_freeze_wt=0.05,**kwargs):
        if portfolio is None:
            raise Exception("mcsamplerPortfolio: must provide portfolio on init")
        self.portfolio=portfolio
        self.portfolio_realizations = []
        self.oracle_realizations= oracle_realizations if oracle_realizations else []
        self.portfolio_member_varaha = {} # these members ONLY train from their own data (i.e., VARAHA), and use likelihood contours (VARAHA)
        self.portfolio_draw_iteration = 0  #  counter, used fo
        self.portfolio_breakpoints = None # breakpoints, at which we activate the other samplers for (a) drawing and (b) training
        for member in self.portfolio:
            if isinstance(member, ModuleType):
              # can pass it a top-level routine, OR
              sampler = member.MCSampler()
              self.portfolio_realizations.append(sampler)
            else:
              # can pass low-level sampler object itself
              self.portfolio_realizations.append(member)

        self.portfolio_weights =portfolio_weights   # cpu-type data structure !
        if not(self.portfolio_weights ):
            self.portfolio_weights = np.ones(len(self.portfolio))/(1.0*len(self.portfolio))

        self.portfolio_adapt = np.ones(len(self.portfolio),dtype=bool) # default : everything adapts.  
        self.portfolio_freeze_wt =portfolio_freeze_wt  # if weight is below this number, the portfolio member's distribution will NOT update. SCALAR

        # Total number of samples drawn
        self.ntotal = 0
        # Parameter names
        self.params = set()
        self.params_ordered = []  # keep them in order. Important to break likelihood function need for names
        # If the pdfs aren't normalized, this will hold the normalization 
        # Cache for the sampling points
        self._rvs = {}
        # parameter -> cdf^{-1} function object
        # params for left and right limits
        self.llim, self.rlim = {}, {}


        self.n_chunk = n_chunk
        self.nbins = None
        self.adaptive =[]


        # histogram setup
        self.xpy = numpy
        self.identity_convert = lambda x: x  # if needed, convert to numpy format  (e.g, cupy.asnumpy)
        self.identity_convert_togpu = lambda x: x  # if needed, convert to numpy format  (e.g, cupy.asnumpy)

        # extra args, created during setup
        self.extra_args = {}

    def add_parameter(self, params, pdf,  **kwargs):
        """
        Add one (or more) parameters to sample dimensions. params is either a string describing the parameter, or a tuple of strings. The tuple will indicate to the sampler that these parameters must be sampled together. left_limit and right_limit are on the infinite interval by default, but can and probably should be specified. If several params are given, left_limit, and right_limit must be a set of tuples with corresponding length. Sampling PDF is required, and if not provided, the cdf inverse function will be determined numerically from the sampling PDF.
        """
        self.params.add(params) # does NOT preserve order in which parameters are provided
        self.params_ordered.append(params)
        for member in self.portfolio_realizations  + self.oracle_realizations:
            member.add_parameter(params, pdf, **kwargs)
        # update dictionary limits
        self.llim.update( member.llim)
        self.rlim.update(member.rlim)
        # set master list of adaptive parameters 
        self.adaptive = member.adaptive  # top level list of adaptive coordinates


    def setup(self,  **kwargs):
        self.extra_args =kwargs  # may need to pass/use during the 'update' step
        if 'oracle_realizations' in kwargs:
          if kwargs['oracle_realizations']: 
            self.oracle_realizations = kwargs['oracle_realizations']  # might not have been initialized earlier
        if (not('portfolio_breakpoints') in kwargs) or not(self.portfolio_breakpoints):
          self.portfolio_breakpoints = np.zeros(len(self.portfolio)) # always use all of them
        if 'portfolio_breakpoints' in kwargs:
            if kwargs['portfolio_breakpoints']:
              self.portfolio_breakpoints =np.array( kwargs['portfolio_breakpoints']     )
        assert len(self.portfolio_breakpoints) == len(self.portfolio_realizations)  # must match


        portfolio_extra_args = [{} for x in self.portfolio_realizations] # empty list
        if 'portfolio_args' in kwargs:
          if not(kwargs['portfolio_args'] is None):
            if len(kwargs['portfolio_args']) == len(self.portfolio_realizations): # Only pass args if valid
              portfolio_extra_args = kwargs['portfolio_args']
            else:
              print(" PORTFOLIO - format ERROR ", kwargs['portfolio_args'])
        for indx, member in enumerate(self.portfolio):
            if hasattr(member, 'setup'):
              print(" PORTFOLIO setup ", member, portfolio_extra_args[indx])
              args_here = {}
              args_here.update(kwargs)
              args_here.update(portfolio_extra_args[indx])
              member.setup(**args_here)
        for indx, member in enumerate(self.oracle_realizations):
            if hasattr(member, 'setup'):
              print(" PORTFOLIO ORACLE setup ", member, portfolio_extra_args[indx])
              args_here = {}
              args_here.update(kwargs)
              args_here.update(portfolio_extra_args[indx])
              member.setup(**args_here)
              member.params_ordered = list(self.params_ordered)  # enforce parameters for oracle being sane

    def draw(self,n_samples, *args, **kwargs):
        """
        draw n_samples

        Draw from portfolio.
        Uses portfolio weights to calculate desired outcomes.
        Restricts to ACTIVE members.
        """
        if len(args) == 0:
            args = self.params
        n_params = len(args)
        n_samples = int(n_samples)

        self.portfolio_draw_iteration += 1


        # Allocate memory.
        #    - initialize with zeros so we will hard fail /nan if error
        rv = self.xpy.empty((n_params, n_samples), dtype=numpy.float64)
        joint_p_s = self.xpy.zeros(n_samples, dtype=numpy.float64)
        joint_p_prior = self.xpy.zeros(n_samples, dtype=numpy.float64)

        indx_active = np.argwhere(self.portfolio_breakpoints <= self.portfolio_draw_iteration).flatten() # provide indexes
        weights_active = np.array([self.portfolio_weights[x] for x in indx_active]) # only provide desired ones
        weights_active *= 1./np.sum(weights_active)  # renormalize
        portfolio_active = [self.portfolio_realizations[x] for x in indx_active] # get the active portfolio members
#        print(" \t ",indx_active, self.portfolio_breakpoints, self.portfolio_draw_iteration)

        # if only one method is active, just call the low-level function
        if len(indx_active) == 1:
           joint_p_s, joint_p_prior, rv = self.portfolio[indx_active[0]].draw_simplified(n_samples, *self.params_ordered, **kwargs)
        else:
          # Identify number of samples per member of the portfolio. Can be zero.
          n_samples_per_member = ((np.array(weights_active))*n_samples).astype(int)

          # logic to block cases where we zero out a number of samples per member.
          # Note this motivates keeping portfolio adaptive weights frozen and not too small, to avoid accidental negative counts.
          if np.sum(n_samples_per_member[0:-1]) < n_samples:
            n_samples_per_member[-1] = n_samples - np.sum(n_samples_per_member[0:-1])
          elif np.sum(n_samples_per_member[0:-2]) < n_samples:
            n_samples_per_member[-1] = 0
            n_samples_per_member[-2] = n_samples - np.sum(n_samples_per_member[0:-2])

          n_index_start_per_member = np.zeros(len(portfolio_active),dtype=int)
          n_index_start_per_member[1:] = np.cumsum(n_samples_per_member)[:-1]

          # Draw in blocks, and copy in place
          # only draw from ACTIVE members
          for indx_member, member in enumerate(portfolio_active):
            joint_p_s_here, joint_p_prior_here, rv_here = member.draw_simplified(
                n_samples_per_member[indx_member], *self.params_ordered, **kwargs
                )
            # type convert as needed, to GPU
            if not(isinstance( type(joint_p_s_here), type(joint_p_s))):
              joint_p_s_here = self.identity_convert_togpu(joint_p_s_here)
              joint_p_prior_here = self.identity_convert_togpu(joint_p_prior_here)
              rv_here = self.identity_convert_togpu(rv_here)
            indx_start = int(n_index_start_per_member[indx_member])
            indx_end = indx_start + int(n_samples_per_member[indx_member])
            joint_p_s[indx_start:indx_end] = joint_p_s_here
            joint_p_prior[indx_start:indx_end] = joint_p_prior_here
            rv[:,indx_start:indx_end] = rv_here
            
        #
        # Cache the samples we chose.  REQUIRED
        #
        if True:
         if len(self._rvs) == 0:
            self._rvs = dict(list(zip(args, rv)))
         else:
            rvs_tmp = dict(list(zip(args, rv)))
            #for p, ar in self._rvs.items():
            for p in self.params_ordered:
                self._rvs[p] = numpy.hstack( (self._rvs[p], rvs_tmp[p]) )


        return joint_p_s, joint_p_prior, rv


    def integrate(self, lnF, *args, xpy=xpy_default,**kwargs):
        use_lnL = kwargs['use_lnL'] if 'use_lnL' in kwargs else False
        if not(use_lnL):
          raise Exception("mcsamplerPortfolio: must integrate lnL")
        return self.integrate_log(lnF, *args, xpy=xpy, **kwargs)
        

    def integrate_log(self, lnF, *args, xpy=xpy_default,**kwargs):
        xpy_here = self.xpy
        
        #
        # Determine stopping conditions
        #
        nmax = kwargs["nmax"] if "nmax" in kwargs else float("inf")
        neff = kwargs["neff"] if "neff" in kwargs else numpy.float128("inf")
        n = int(kwargs["n"] if "n" in kwargs else min(100000, nmax))
        convergence_tests = kwargs["convergence_tests"] if "convergence_tests" in kwargs else None
        save_no_samples = kwargs["save_no_samples"] if "save_no_samples" in kwargs else None
        portfolio_wt_func = kwargs['portfolio_schedule'] if 'portfolio_schedule' in kwargs else portfolio_default_weights

        #
        # Adaptive sampling parameters
        #
        n_history = int(kwargs["history_mult"]*n) if "history_mult" in kwargs else 2*n
        if n_history<=0:
            print("  Note: cannot adapt, no history ")

        tempering_exp = kwargs["tempering_exp"] if "tempering_exp" in kwargs else 0.0
        n_adapt = int(kwargs["n_adapt"]*n) if "n_adapt" in kwargs else 1000  # default to adapt to 1000 chunks, then freeze
        floor_integrated_probability = kwargs["floor_level"] if "floor_level" in kwargs else 0
        temper_log = kwargs["tempering_log"] if "tempering_log" in kwargs else False
        tempering_adapt = kwargs["tempering_adapt"] if "tempering_adapt" in kwargs else False
            

        save_intg = kwargs["save_intg"] if "save_intg" in kwargs else False
        # FIXME: The adaptive step relies on the _rvs cache, so this has to be
        # on in order to work
        if n_adapt > 0 and tempering_exp > 0.0:
            save_intg = True

        deltalnL = kwargs['igrand_threshold_deltalnL'] if 'igrand_threshold_deltalnL' in kwargs else float("Inf") # default is to return all
        deltaP    = kwargs["igrand_threshold_p"] if 'igrand_threshold_p' in kwargs else 0 # default is to omit 1e-7 of probability
        bFairdraw  = kwargs["igrand_fairdraw_samples"] if "igrand_fairdraw_samples" in kwargs else False
        n_extr = kwargs["igrand_fairdraw_samples_max"] if "igrand_fairdraw_samples_max" in kwargs else None

        bShowEvaluationLog = kwargs['verbose'] if 'verbose' in kwargs else False
        bShowEveryEvaluation = kwargs['extremely_verbose'] if 'extremely_verbose' in kwargs else False


        verbose = kwargs["verbose"] if "verbose" in kwargs else False  # default
        super_verbose = kwargs["super_verbose"] if "super_verbose" in kwargs else False  # default
        dict_return_q = kwargs["dict_return"] if "dict_return" in kwargs else False  # default.  Method for passing back rich data structures for debugging

        # use integer numbers of bins always
        enforce_bounds = kwargs["enforce_bounds"] if "enforce_bounds" in kwargs else False

        if bShowEvaluationLog:
            print(" .... mcsampler : providing verbose output ..... ")

        current_log_aggregate = None
        eff_samp = 0  # ratio of max weight to sum of weights
        maxlnL = -np.inf  # max lnL
        maxval=0   # max weight
        outvals=None  # define in top level scope
        self.ntotal = 0
        if bShowEvaluationLog:
            print("iteration Neff  sqrt(2*lnLmax) sqrt(2*lnLmarg) ln(Z/Lmax) int_var")

        self.n_chunk = n
#        self.setup()  # sets up self.my_ranges, self.dx initially


        n_zero_prior =0
        it_max_oracle = 7
        it_now  =0
        if 'integrand' in self._rvs:
          # remove conflict
          del self._rvs['integrand']
        while (eff_samp < neff and self.ntotal < nmax): #  and (not bConvergenceTests):
            

            ###
            ### COMMON INTEGRATION BLOCK
            ###

            # Draw our sample points
            # non-log draw
            joint_p_s, joint_p_prior, rv = self.draw(
                n, *self.params_ordered
            )
            it_now +=1 

            #
            # Unpack rvs and evaluate integrand
            #
            if len(rv[0].shape) != 1:
                rv = rv[0]

            params = []
            for item in self.params_ordered:  # USE IN ORDER
                if isinstance(item, tuple):
                    params.extend(item)
                else:
                    params.append(item)
            unpacked = unpacked0 = rv #numpy.hstack([r.flatten() for r in rv]).reshape(len(args), -1)
            unpacked = dict(list(zip(params, unpacked)))

            # Evaluate function, protecting argument order
            if 'no_protect_names' in kwargs:
                lnL = lnF(*unpacked0)  # do not protect order
            else:
                lnL= lnF(**unpacked)  # protect order using dictionary
            # take log if we are NOT using lnL
            if cupy_ok:
              if not(isinstance(lnL,cupy.ndarray)):
                lnL = identity_convert_togpu(lnL)  # send to GPU, if not already there

            log_integrand =lnL + self.xpy.log(joint_p_prior) - self.xpy.log(joint_p_s)
            # tempering_exp done inside the update proposal, NOT here
            log_weights = lnL + self.xpy.log(joint_p_prior) - self.xpy.log(joint_p_s)

            if save_intg:
                # FIXME: See warning at beginning of function. The prior values
                # need to be moved out of this, as they are not part of MC
                # integration
                if "log_integrand" in self._rvs:
                    self._rvs["log_integrand"] = xpy_here.hstack( (self._rvs["log_integrand"], lnL) )
                    self._rvs["log_joint_prior"] = xpy_here.hstack( (self._rvs["log_joint_prior"], self.xpy.log(joint_p_prior)) )
                    self._rvs["log_joint_s_prior"] = xpy_here.hstack( (self._rvs["log_joint_s_prior"], self.xpy.log(joint_p_s)))
                    self._rvs["log_weights"] = xpy_here.hstack( (self._rvs["log_weights"], log_weights ))
                else:
                    self._rvs["log_integrand"] = lnL
                    self._rvs["log_joint_prior"] = self.xpy.log(joint_p_prior)
                    self._rvs["log_joint_s_prior"] = self.xpy.log(joint_p_s)
                    self._rvs["log_weights"] = log_weights
            # maxlnL
            maxlnL_now = identity_convert(xpy.max(lnL))
            maxlnL = identity_convert(maxlnL)
            if np.isinf(maxlnL ):
              maxlnL = maxlnL_now
            else:
              maxlnL = np.max([maxlnL, maxlnL_now,-100])


            # n, Mean, error tracked by statutils structure
            if current_log_aggregate is None:
              current_log_aggregate = init_log(log_integrand,xpy=xpy,special=xpy_special_default)
            else:
              current_log_aggregate = update_log(current_log_aggregate, log_integrand,xpy=xpy,special=xpy_special_default)
            outvals = finalize_log(current_log_aggregate,xpy=xpy)
            self.ntotal = current_log_aggregate[0]
            # effective samples
            maxval = max(maxval, identity_convert(self.xpy.max(log_integrand) ))

            # sum of weights is the integral * the number of points
            eff_samp = xpy.exp(  outvals[0]+np.log(self.ntotal) - maxval)   # integral value minus floating point, which is maximum


            # Throw exception if we get infinity or nan
            if math.isnan(eff_samp):
                raise NanOrInf("Effective samples = nan")

            if bShowEvaluationLog:
                print(" :",  self.ntotal, eff_samp, numpy.sqrt(2*maxlnL), numpy.sqrt(2*outvals[0]), outvals[0]-maxlnL, np.exp(outvals[1]/2  - outvals[0]  - np.log(self.ntotal)/2 ))

            if (not convergence_tests) and self.ntotal >= nmax and neff != float("inf"):
                print("WARNING: User requested maximum number of samples reached... bailing.", file=sys.stderr)


            if self.ntotal > n_adapt*n:
                print(n_adapt,self.n_total)
                continue

            ###
            ### PORTFOLIO REPORT BLOCK (and reweighting of member priority)
            ###
            #  n_ess for each portfolio member
            #  computed with TRUE weights, not 
            #  Use this to reassess which portfolio members are being refined.
            n_samples = len(log_weights)
            n_samples_per_member = ((self.portfolio_weights)*len(log_weights)).astype(int)
            if np.sum(n_samples_per_member[0:-1]) < n_samples:
              n_samples_per_member[-1] = n_samples - np.sum(n_samples_per_member[0:-1])
            elif np.sum(n_samples_per_member[0:-2]) < n_samples:
              n_samples_per_member[-1] = 0
              n_samples_per_member[-2] = n_samples - np.sum(n_samples_per_member[0:-2])

            n_index_start_per_member = np.zeros(len(self.portfolio_realizations),dtype=int)
            n_index_start_per_member[1:] = np.cumsum(n_samples_per_member)[:-1]

            portfolio_report = {}
            for indx_member, member in enumerate(self.portfolio):
              indx_start = int(n_index_start_per_member[indx_member])
              indx_end = indx_start + int(n_samples_per_member[indx_member])    
              ln_wt_here =  log_weights[indx_start:indx_end]
              ln_wt_here += - np.max(ln_wt_here)
              # evaluate  n_ess, n_eff for this set of samples in batch specifically,
              portfolio_report[indx_member] = [ self.portfolio_weights[indx_member], self.identity_convert(self.xpy.sum(self.xpy.exp(ln_wt_here))**2/self.xpy.sum(self.xpy.exp(ln_wt_here*2))), identity_convert(self.xpy.sum(self.xpy.exp(ln_wt_here)))]
            print("\t",portfolio_report)
            # Weight based on n_ESS from batch.  remember these are >=1, so no negatives or 0 will happen
            dat =np.array([ portfolio_report[k][1] for k in range(len(self.portfolio))])
            self.portfolio_weights = portfolio_wt_func(dat, self.portfolio_weights, xpy=self.xpy, identity_convert=self.identity_convert) # call weighting function

              
            ###
            ### ORACLE BLOCK
            ###
            rvs_train = self._rvs
            if it_now < it_max_oracle and len(self.oracle_realizations )>0:
              rvs_train = deepcopy(self._rvs)  # duplicate deeply, since we will append to it
              n_samples_per_oracle = int(n*0.1/len(self.oracle_realizations)) # try to minimize oracle effort
              print(" ORACLE: attempting updates ")
              # update each oracle
              for member in self.oracle_realizations:
                member.update_sampling_prior(log_weights, n_history, external_rvs=rvs_train, log_scale_weights=True)
              # generate samples from oracles
              rv_oracle = self.xpy.empty((n_samples_per_oracle*len(self.oracle_realizations), len(self.params_ordered)))
              base_now = 0
              for member in self.oracle_realizations:
                _, _, rv_here = member.draw_simplified(n_samples_per_oracle)
                rv_oracle[base_now:base_now+n_samples_per_oracle] = rv_here
                base_now += n_samples_per_oracle
              # evaluate lnL for each,
              lnL_oracles = lnF(*rv_oracle.T)
              # put into weights and rvs, for use in training other samples
              self.xpy.append(log_weights, lnL_oracles)
              for indx, p in enumerate(self.params_ordered):
                self.xpy.append(rvs_train[p],  rv_oracle[:,indx])


            ###
            ### WEIGHT UPDATE BLOCK (improve by adding all portfolio options - default vanilla independent updates now)
            ###
            update_dict = {}
            update_dict.update(self.extra_args)
            update_dict['tempering_exp'] =tempering_exp
            for indx, member in enumerate(self.portfolio_realizations):
                # update sampling prior, using ALL past data
                # Don't update samples which are not being drawn
                # always update if we have an oracle  - don't freeze out out oracle, UNLESS we have explicitly frozen it with a breakpoint
                if self.portfolio_draw_iteration < self.portfolio_breakpoints[indx]:
                  print("  - before activation breakpoint for member {} ".format( indx))
                  pass
                elif (len(self.oracle_realizations) > 0 and it_now <it_max_oracle) or (   self.portfolio_weights[indx] > self.portfolio_freeze_wt):
                  if not(hasattr(member, 'is_varaha')):
                    member.update_sampling_prior(log_weights, n_history,external_rvs=rvs_train,log_scale_weights=True, **update_dict)
                  else:
                    # just do a single VARAHA step, independent of others
                    member.update_sampling_prior_selfish(lnF)
                else:
                  if self.portfolio_draw_iteration > self.portfolio_breakpoints[indx]:  
                    print("   - frozen sampling for member {} {}".format(indx, self.portfolio_weights[indx]))
                  else:
                    print("  - before activation breakpoint for member {} ".format( indx))

        # If we were pinning any values, undo the changes we did before
        # self.pdf.update(temppdfdict)
        # self._pdf_norm.update(temppdfnormdict)
        # self.prior_pdf.update(temppriordict)

        # Clean out the _rvs arrays for 'irrelevant' points
        #   - find and remove samples with  lnL less than maxlnL - deltalnL (latter user-specified)
        #   - create the cumulative weights
        #   - find and remove samples which contribute too little to the cumulative weights
        if (not save_no_samples) and ( "log_integrand" in self._rvs):
            self._rvs["sample_n"] = self.identity_convert_togpu(numpy.arange(len(self._rvs["log_integrand"])))  # create 'iteration number'        
            # Step 1: Cut out any sample with lnL belw threshold
            if deltalnL < 1e10: # not infinity, so we are truncating the sample list
              indx_list = [k for k, value in enumerate( (self._rvs["log_integrand"] > maxlnL - deltalnL)) if value] # threshold number 1
              # FIXME: This is an unncessary initial copy, the second step (cum i
              # prob) can be accomplished with indexing first then only pare at
              # the end
              for key in list(self._rvs.keys()):
                if isinstance(key, tuple):
                    self._rvs[key] = self._rvs[key][:,indx_list]
                else:
                    self._rvs[key] = self._rvs[key][indx_list]
            # Step 2: Create and sort the cumulative weights, among the remaining points, then use that as a threshold
            ln_wt = self._rvs["log_integrand"] + self._rvs["log_joint_prior"] - self._rvs["log_joint_s_prior"]
            # Convert to CPU as needed
            ln_wt = identity_convert(ln_wt)
            ln_wt += - np.max(ln_wt)  # remove maximum value, irrelevant
            wt = np.exp(ln_wt) # exponentiate.  Danger underflow
            idx_sorted_index = numpy.lexsort((numpy.arange(len(wt)), wt))  # Sort the array of weights, recovering index values
            indx_list = numpy.array( [[k, ln_wt[k]] for k in idx_sorted_index])     # pair up with the weights again. NOTE NOT INTEGER TYPE ANY MORE
            cum_sum = numpy.cumsum(indx_list[:,1])  # find the cumulative sum
            cum_sum = cum_sum/cum_sum[-1]          # normalize the cumulative sum
            indx_list = [int(indx_list[k, 0]) for k, value in enumerate(cum_sum > deltaP) if value]  # find the indices that preserve > 1e-7 of total probability. RECAST TO INTEGER
            # FIXME: See previous FIXME
            for key in list(self._rvs.keys()):
                if isinstance(key, tuple):
                    self._rvs[key] = self._rvs[key][:,indx_list]
                else:
                    self._rvs[key] = self._rvs[key][indx_list]

        # Do a fair draw of points, if option is set. CAST POINTS BACK TO NUMPY, IDEALLY
        if bFairdraw and not(n_extr is None):
           n_extr = int(numpy.min([n_extr,1.5*identity_convert(eff_samp),1.5*neff]))
           print(" Fairdraw size : ", n_extr)
           ln_wt = self.xpy.array(self._rvs["log_integrand"] + self._rvs["log_joint_prior"] - self._rvs["log_joint_s_prior"] ,dtype=float)
           ln_wt = identity_convert(ln_wt)  # send to CPU
           ln_wt += - special.logsumexp(ln_wt)
           wt = xpy.exp(identity_convert_togpu(ln_wt))
           if n_extr < len(self._rvs["log_integrand"]):
               indx_list = self.xpy.random.choice(self.xpy.arange(len(wt)), size=n_extr,replace=True,p=wt) # fair draw
               # FIXME: See previous FIXME
               for key in list(self._rvs.keys()):
                   if isinstance(key, tuple):
                       self._rvs[key] = identity_convert(self._rvs[key][:,indx_list])
                   else:
                       self._rvs[key] = identity_convert(self._rvs[key][indx_list])


        # Create extra dictionary to return things
        dict_return ={}
        # if convergence_tests is not None:
        #     dict_return["convergence_test_results"] = None # last_convergence_test

        # perform type conversion of all stored variables
        if cupy_ok:
          for name in self._rvs:
            if isinstance(self._rvs[name],xpy_default.ndarray):
              self._rvs[name] = identity_convert(self._rvs[name])   # this is trivial if xpy_default is numpy, and a conversion otherwise

        # Return.  Take care of typing
        if outvals:
          out0 = outvals[0]; out1 = outvals[1]
          if not(isinstance(outvals[0], np.float64)):
            # type convert everything as needed
            out0 = identity_convert(out0)
          if not(isinstance(outvals[1], np.float64)):
            out1 = identity_convert(out1)
            eff_samp = identity_convert(eff_samp)
          self._rvs['integrand'] = self._rvs['log_integrand'] # always integrating log function.  Match behavior of other routines
          return out0, out1 - np.log(self.ntotal), eff_samp, dict_return
        else: # very strange case where we terminate early
          return None, None, None, None
